#include <type.h>
#include <kernel/kernel.h>

static object *_clones;

int send_message(mixed message) {
  return ::send_message(message);
}

int is_clone() {
  return (sscanf(object_name(this_object()), "%*s#%*s")==2);
}

string _mixed_to_string(string spaces, mixed var) {
  mixed *keys;
  mixed *values;
  int i;
  string result;

  if (!var) {
    result=spaces+"null";
  }
  switch (typeof(var)) {
    case T_NIL:
      result=spaces+"nil";
      break;
    case T_INT:
      result=spaces+"int:"+var;
      break;
    case T_FLOAT:
      result=spaces+"float:"+var;
      break;
    case T_STRING:
      result=spaces+"string:"+var;
      break;
    case T_OBJECT:
      result=spaces+"object:"+object_name(var);
      break;
    case T_ARRAY:
      result=spaces+"array:({";
      for (i=0;i<sizeof(var);i++) {
        result+="\n"+_mixed_to_string(spaces+"  ", var[i]);
      }
      result+="})";
      break;
    case T_MAPPING:
      result=spaces+"mapping:([";
      keys=map_indices(var);
      values=map_values(var);
      for (i=0;i<map_sizeof(var);i++) {
        result+="\n"+_mixed_to_string(spaces+"  ", keys[i])+"::"+_mixed_to_string(spaces+"  ", values[i]);
      }
      result+="])";
      break;
    default:
      result=spaces+"unknown";
  }
  return result;
}

string mixed_to_string(mixed var) {
  return _mixed_to_string("", var);
}

void debug_message(mixed message) {
  string objname;

  if (this_object()) {
    objname=object_name(this_object())+":";
  } else {
    objname="nothing:";
  }
  if (typeof(message)==T_STRING) {
    write_file("/log/debug", objname+message+"\n");
    DRIVER->send_message(objname+message+"\n");
    if (this_user()) {
      this_user()->send_message(objname+message+"\n");
    }
  } else {
    write_file("/log/debug", objname+mixed_to_string(message)+"\n");
    DRIVER->send_message(objname+mixed_to_string(message)+"\n");
    if (this_user()) {
      this_user()->send_message(objname+mixed_to_string(message)+"\n");
    }
  }
}

void create() {
}

void _create() {
  /* create() gets called for clones of objects in a .../obj/... directory, and
     for blueprints of objects that are not in a .../obj/... or .../lib/...
     directory, but not for anything else.  In particular, inherited objects
     and the blueprints of clones don't get create() called. */
  string name;

  name=object_name(this_object());
  /* It can only be a clone if it was in a .../obj/... directory, so no need
     to explicitly check for that */
  if (sscanf(name, "%*s#")) {
    this_object()->create();
  } else {
    if (!sscanf(name, "%*s/lib/") && !sscanf(name, "%*s/obj/")) {
      this_object()->create();
    }
  }
}

void send_event(mixed targets, string event_name, mixed args...) {
  string function_name;
  object *t;

  if (typeof(targets)!=T_ARRAY) {
    targets=({targets});
  }
  t=({ })+targets;
  function_name="event_"+event_name;
  while (sizeof(t)) {
    call_other(t[0], function_name, this_object(), args...);
    t-=({ t[0] });
  }
}

string lower_case(string s) {
  int i;

  for (i=0;i<strlen(s);i++) {
    if (s[i]<='Z' && s[i]>='A') {
      s[i]=s[i]-'A'+'a';
    }
  }
  return s;
}

string upper_case(string s) {
  int i;

  for (i=0;i<strlen(s);i++) {
    if (s[i]<='z' && s[i]>='a') {
      s[i]=s[i]-'a'+'A';
    }
  }
  return s;
}

object compile_object(string name) {
  object o;

  o=::compile_object(name);
  /* Make sure create() gets called */
  call_other(o, "???");
  DRIVER->learn_inheritence(name, "/kernel/lib/auto");
  return o;
}

object find_object(string path) {
  object obj;

  if (lower_case(path[strlen(path)-2..strlen(path)-1])==".c") {
    path=path[..strlen(path)-3];
  }
  obj=::find_object(path);
  if (!obj) {
    obj=compile_object(path);
  }
  return obj;
}

object blueprint(object obj) {
  return find_object(explode(object_name(obj), "\n")[0]);
}

void _add_clone(object obj) {
  if (!_clones) {
    _clones=({ });
  }
  _clones+=({obj});
}

void _remove_clone(object obj) {
  if (!_clones) {
    _clones=({ });
  }
  _clones-=({obj});
}

object *query_clones() {
  return _clones;
}

void destruct_object(object obj) {
  /* If it is a physical object, we have to remove it from the universe first */
  if (obj) {
    obj->remove_from_universe();
    obj->prepare_to_die();
    blueprint(obj)->_remove_clone(obj);
    ::destruct_object(obj);
  }
}

void input_to(string function) {
  ::this_user()->_input_to(this_object(), function, 0);
}

void input_to_no_echo(string function) {
  ::this_user()->_input_to(this_object(), function, 1);
}

object clone_object(object obj) {
  object result;

  if (sscanf(object_name(obj), "%*s/lib/")) {
    DRIVER->message("Illegal clone: "+ mixed_to_string(this_object())+
      " attempting to clone "+mixed_to_string(obj));
    return nil;
  }
  if (!sscanf(object_name(obj), "%*s/obj/")) {
    DRIVER->message("Illegal clone: "+ mixed_to_string(this_object())+
      " attempting to clone "+mixed_to_string(obj));
    return nil;
  }
  result=::clone_object(obj);
  obj->_add_clone(result);
  return result;
}

void depends_on(mixed obj) {
  if (typeof(obj)==T_OBJECT) {
    DRIVER->learn_inheritence(object_name(this_object()), object_name(obj));
  } else {
    DRIVER->learn_inheritence(object_name(this_object()), obj);
  }
}

object this_user() {
  /* ::this_user() returns the connection object, which is not what we want */
  return USER_MANAGER->query_user(::this_user());
}

object *users() {
  return USER_MANAGER->query_users();
}

int abs(int x) {
  if (x<0) {
    return -x;
  } else {
    return x;
  }
}
