#include <kernel/universe.h>
#include <kernel/parser.h>

#define ONE_OBJECT 0
#define MANY_OBJECTS 1

inherit "/kernel/lib/strings";

/* grammar is an array of grammar lines.  Each line contains:
   0: The object that defined it (object)
   1: The verb (string)
   2: The pattern (array of strings)
   3: The handler function (string)
   4: The function parameters (array of integers) 
   5: Invisibility flag (1 means it won't show up in a list of commands)

   The pattern contains constant words and tokens.  
   The handler function will be passed the verb, then object arrays for each
   item listed in the function parameters, then strings for each item in the
   function parameters.
   The function parameters specifier is an array of word numbers in the pattern,
   with 0 being the first word.  An index of -1 is a blank - the function will
   be passed an empty object array and an empty string in this position.
*/
static mixed **grammar;

void add_grammar(string verb, string pattern, string function, int *params, varargs int invisible) {
  if (!grammar) {
    grammar=({ });
  }
  grammar+=({({previous_object(), verb, explode(pattern, " "), function, params, invisible})});
}

mixed *query_grammar() {
  return grammar;
}

void remove_command(object obj, string verb) {
/* Remove all grammar lines for this verb that were added by the object */
  mixed **remove;
  int i;

  /* Build a list of everything to be removed, and remove it in one step
     - safer than removing things from a list as we're walking through it */
  remove=({ });
  for (i=0;i<sizeof(grammar);i++) {
    if (grammar[i][GRAMMAR_OBJ]==obj && grammar[i][GRAMMAR_VERB]==verb) {
      remove+=({grammar[i]});
    }
  }
  grammar-=remove;
}

void remove_grammar(object obj) {
  /* Remove all grammar lines added by the object */
  mixed **remove;
  int i;

  /* Build a list of everything to be removed, and remove it in one step
     - safer than removing things from a list as we're walking through it */
  remove=({ });
  for (i=0;i<sizeof(grammar);i++) {
    if (grammar[i][GRAMMAR_OBJ]==obj) {
      remove+=({grammar[i]});
    }
  }
  grammar-=remove;
}

string ***one_match(string *words, string *pattern) {
  /* Return a list of all possible matches of words against pattern */
  string ***rest;
  string ***result;
  int i, j;

  result=({ });
  if (pattern[0][0..0]=="%") {
    /* The first item in pattern is %something */
    for (j=0;j<sizeof(words);j++) {
      if (sizeof(words)>j+1 && sizeof(pattern)>1) {
        rest=one_match(words[j+1..], pattern[1..]);
        if (sizeof(rest)) {
          for (i=0;i<sizeof(rest);i++) {
            result+=({ ({ ({pattern[0], implode(words[0..j], " ") }) })+rest[i]});
          }
        }
      } else {
        if (sizeof(words)==(j+1) && sizeof(pattern)==1) {
          result+=({ ({ ({pattern[0], implode(words[0..j], " ") }) }) });
        }
      }
    }
  } else {
    /* The first item in pattern is a fixed word */
    if (words[0]==pattern[0]) {
      /* First word matches.  Is there any more to match? */
      if (sizeof(words)>1 && sizeof(pattern)>1) {
        /* words and pattern have more. */
        rest=one_match(words[1..], pattern[1..]);
        if (sizeof(rest)) {
          for (i=0;i<sizeof(rest);i++) {
            result+=({ ({ ({pattern[0], words[0]}) })+rest[i]});
          }
        } else {
          result=({ });
        }
      } else {
        /* Either words or pattern have no more.  If neither do, we have a
           match.  If either has more, we have no match */
        if (sizeof(words)>1 || sizeof(pattern)>1) {
          result=({ });
        } else {
          result=({ ({pattern[0], words[0]}) });
        }
      }
    }
  }
  return result;
}

mixed *find_text_matches(string *words) {
  /* Returns a array of ({ grammar_line_number, correspondences })
     correspondences is an array of ({ token, string }). */
  int i, j;
  mixed *matches;
  mixed *result;
  string verb;

  if (!words) {
    return ({ });
  }
  verb=words[0];
  words=words[1..];
  result=({ });
  for (i=sizeof(grammar)-1;i>=0;i--) {
    if (grammar[i][GRAMMAR_VERB]==verb) {
      if (sizeof(words)==0) {
        if (sizeof(grammar[i][GRAMMAR_PATTERN])==0) {
          result+=({ ({ i, ({ }) }) });
        }
      } else {
        if (sizeof(grammar[i][GRAMMAR_PATTERN])) {
          matches=one_match(words, grammar[i][GRAMMAR_PATTERN]);
          for (j=0;j<sizeof(matches);j++) {
            result+=({ ({ i, matches[j] }) });
          }
        }
      }
    }
  }
  return result;
}

mixed get_objs_from_candidates(string text, object *candidates, object *scope, int number) {
  /* Return an array containing a score and a list of objects that match 'text'.
     The objects are chosen from 'candidates'.  'scope' is the list of all
     other objects that might be required - for example, if 'text' is "fish in
     bowl", we should only be looking for objects that match "fish" that are in
     objects from 'scope' that match "bowl".  'number' determines whether we're
     expecting to match one object or many. */
  object *result;
  mixed *contresult;
  mixed *objresult;
  mixed *res1;
  mixed *res2;
  string rest, s1, s2;
  int i;

  result=({ });
  if (sscanf(text, "the %s", rest)) {
    text=rest;
  } else if (sscanf(text, "a %s", rest)) {
    text=rest;
  } else if (sscanf(text, "an %s", rest)) {
    text=rest;
  }
  if (sscanf(text, "%s except %s", s1, s2)==2) {
    res1=get_objs_from_candidates(s1, candidates, scope, number);
    res2=get_objs_from_candidates(s2, candidates, scope, number);
    if (res1[0]>=0 && (res2[0]>=0)) {
      return ({ res1[0]+res2[0], res1[1]-res2[1] });
    } else {
      return ({ -1000, ({ }) });
    }
  } else if (sscanf(text, "%s,%s", s1, s2)==2) {
    res1=get_objs_from_candidates(s1, candidates, scope, number);
    res2=get_objs_from_candidates(s2, candidates, scope, number);
    if (res1[0]>=0 && (res2[0]>=0)) {
      return ({ res1[0]+res2[0], res1[1]+res2[1] });
    } else {
      return ({ -1000, ({ }) });
    }
  } else if (sscanf(text, "%s, %s", s1, s2)==2) {
    res1=get_objs_from_candidates(s1, candidates, scope, number);
    res2=get_objs_from_candidates(s2, candidates, scope, number);
    if (res1[0]>=0 && (res2[0]>=0)) {
      return ({ res1[0]+res2[0], res1[1]+res2[1] });
    } else {
      return ({ -1000, ({ }) });
    }
  } else if (sscanf(text, "%s and %s", s1, s2)==2) {
    res1=get_objs_from_candidates(s1, candidates, scope, number);
    res2=get_objs_from_candidates(s2, candidates, scope, number);
    if (res1[0]>=0 && (res2[0]>=0)) {
      return ({ res1[0]+res2[0], res1[1]+res2[1] });
    } else {
      return ({ -1000, ({ }) });
    }
  } else if (sscanf(text, "%s in %s", s1, s2)==2) {
    contresult=get_objs_from_candidates(s2, scope, scope, ONE_OBJECT);
    if (contresult[0]>=0) {
      /* We've got a list of potential containers now.  For each one, call
         get_objs_from_candidates, with candidates=the (shallow) inventory
         of the container. */
      for (i=0;i<sizeof(contresult[1]);i++) {
        objresult=get_objs_from_candidates(s1, contresult[1][i]->query_shallow_inventory(), contresult[1][i]->query_shallow_inventory(), number);
        if (objresult[0]>=0) {
          return objresult;
        }
      }
      /* We've searched all containers, and found nothing.  Fail */
      return ({ -1000, ({ }) });
    } else {
      return ({ -1000, ({ }) });
    }
  }
  if (text=="all") {
    result=candidates;
  } else if (text=="here") {
    result=({ this_object()->query_environment() });
  } else if (text=="me") {
    result=({ this_object() });
  } else {
    for (i=0;i<sizeof(candidates);i++) {
      if (candidates[i]->check_id(text)) {
        result+=({ candidates[i] });
      }
    }
  }
  if (sizeof(result)) {
    if (number==ONE_OBJECT) {
      if (sizeof(result)==1) {
        /* If too many match, return the first but indicate that it's not
           the best solution. */
        return ({ 0, ({ result[0] }) });
      } else {
        return ({ 1, result });
      }
    } else {
      return ({ 1, result });
    }
  } else {
    return ({ -1000, ({ }) });
  }
}

mixed get_objs_from_thisobject(object actor, string text, mixed *grammar_line) {
  if (grammar_line[GRAMMAR_OBJ]->check_id(text)) {
    return ({ 1, ({ grammar_line[GRAMMAR_OBJ] }) });
  } else {
    return ({ -1000, ({ }) });
  }
}

mixed get_objs_from_object(object actor, string text, mixed *grammar_line) {
  return get_objs_from_candidates(text, this_object()->near_objects(), this_object()->near_objects(), ONE_OBJECT);
}

mixed get_objs_from_person(object actor, string text, mixed *grammar_line) {
  object *people;
  object *scope;
  int i;

  people=({ });
  scope=this_object()->near_objects();
  for (i=0;i<sizeof(scope);i++) {
    if (scope[i]->query_property(UNIV_PROP_PERSON)) {
      people+=({ scope[i] });
    }
  }
  return get_objs_from_candidates(text, people, scope, ONE_OBJECT);
}

mixed get_objs_from_globalperson(object actor, string text, mixed *grammar_line) {
  object *scope;

  scope=users();
  return get_objs_from_candidates(text, scope, this_object()->all_objects(), ONE_OBJECT);
}

mixed get_objs_from_container(object actor, string text, mixed *grammar_line) {
  object *containers;
  object *scope;
  int i;

  containers=({ });
  scope=this_object()->near_objects();
  for (i=0;i<sizeof(scope);i++) {
    if (scope[i]->query_property(UNIV_PROP_CONTAINER)) {
      containers+=({ scope[i] });
    }
  }
  return get_objs_from_candidates(text, containers, scope, ONE_OBJECT);
}

mixed get_objs_from_inside(object actor, string text, mixed *grammar_line) {
  object *scope;
  object *near;
  object *inv;
  int i;

  near=this_object()->near_objects();
  scope=({ });
  for (i=0;i<sizeof(near);i++) {
    inv=near[i]->shallow_inventory();
    if (inv) {
      scope+=inv;
    }
  }
  return get_objs_from_candidates(text, scope, this_object()->near_objects(), ONE_OBJECT);
}

mixed get_objs_from_outside(object actor, string text, mixed *grammar_line) {
  return get_objs_from_candidates(text, this_object()->near_objects(), this_object()->near_objects(), ONE_OBJECT);
}

mixed get_objs_from_held(object actor, string text, mixed *grammar_line) {
  return get_objs_from_candidates(text, this_object()->query_deep_inventory()-({this_object()}), this_object()->query_deep_inventory(), ONE_OBJECT);
}

mixed get_objs_from_notheld(object actor, string text, mixed *grammar_line) {
  return get_objs_from_candidates(text, this_object()->near_objects()-this_object()->query_deep_inventory()-({this_object()}), this_object()->near_objects(), ONE_OBJECT);
}

mixed get_objs_from_multi(object actor, string text, mixed *grammar_line) {
  return get_objs_from_candidates(text, this_object()->near_objects(), this_object()->near_objects(), MANY_OBJECTS);
}

mixed get_objs_from_multiperson(object actor, string text, mixed *grammar_line) {
  object *people;
  object *scope;
  int i;

  people=({ });
  scope=this_object()->near_objects();
  for (i=0;i<sizeof(scope);i++) {
    if (scope[i]->query_property(UNIV_PROP_PERSON)) {
      people+=({ scope[i] });
    }
  }
  return get_objs_from_candidates(text, people, scope, MANY_OBJECTS);
}

mixed get_objs_from_multiglobalperson(object actor, string text, mixed *grammar_line) {
  object *scope;

  scope=users();
  return get_objs_from_candidates(text, scope, this_object()->all_objects(), MANY_OBJECTS);
}

mixed get_objs_from_multiinside(object actor, string text, mixed *grammar_line) {
  object *scope;
  object *near;
  object *inv;
  int i;

  near=this_object()->near_objects();
  scope=({ });
  for (i=0;i<sizeof(near);i++) {
    inv=near[i]->shallow_inventory();
    if (inv) {
      scope+=inv;
    }
  }
  return get_objs_from_candidates(text, scope, this_object()->near_objects(), MANY_OBJECTS);
}

mixed get_objs_from_multioutside(object actor, string text, mixed *grammar_line) {
  return get_objs_from_candidates(text, this_object()->near_objects(), this_object()->near_objects(), MANY_OBJECTS);
}

mixed get_objs_from_multiheld(object actor, string text, mixed *grammar_line) {
  return get_objs_from_candidates(text, this_object()->query_deep_inventory()-({this_object()}), this_object()->query_deep_inventory(), MANY_OBJECTS);
}

mixed get_objs_from_multinotheld(object actor, string text, mixed *grammar_line) {
  return get_objs_from_candidates(text, this_object()->near_objects()-this_object()->query_deep_inventory()-({this_object()}), this_object()->near_objects(), MANY_OBJECTS);
}

mixed get_objs_from_file(object actor, string text, mixed *grammar_line) {
  string path;
  mixed *file_info;

  path=this_object()->real_dir(text);
  file_info=get_dir(path);
  if (sizeof(file_info[0])) {
    return ({ 1, ({ }) });
  } else {
    file_info=get_dir(path+".c");
    if (sizeof(file_info[0])) {
      return ({ 1, ({ }) });
    } else {
      return ({ -1000, ({ }) });
    }
  }
}

mixed get_objs_from_multifile(object actor, string text, mixed *grammar_line) {
  /* !!! Going to have to do this properly at some point */
  return get_objs_from_file(actor, text, grammar_line);
}

mixed *get_objs_from_word(object actor, string text, mixed *grammar_line) {
  if (sizeof(explode(text, " "))==1) {
    return ({ 1, ({ }) });
  } else {
    return ({ -1000, ({ }) });
  }
}

mixed *get_objs_from_text(object actor, string text, mixed *grammar_line) {
  return ({ 1, ({ }) });
}

mixed *get_objs_from_token(object actor, object obj, string token, string text, mixed *grammar_line) {
  mixed *result;

  if (token[0..0]=="%") {
    if (function_object("get_objs_from_"+token[1..], this_object())) {
      result=call_other(this_object(), "get_objs_from_"+token[1..], actor, text, grammar_line);
    } else {
      result=call_other(obj, "get_objs_from_"+token[1..], actor, text, grammar_line);
    }
  } else {
    result=({ 0, ({ }) });
  }
  return result;
}

mixed *find_objects(object actor, mixed *matches) {
  /* Returns an array of ({ input_number, score, correspondences })
     input_number is the index into *matches that this entry comes from.
     correspondences is an array of ({ token, *objects, text }). */
  int i, j, k;
  int good;
  int score;
  mixed *response;
  mixed *result;
  mixed *this_result;
  object obj;
  object *container;
  object *cont_objs;

  result=({ });
  for (i=0;i<sizeof(matches);i++) {  /* for each match */
    obj=grammar[matches[i][0]][GRAMMAR_OBJ];
    good=1;
    score=0;
    this_result=({ });
    for (j=0;j<sizeof(matches[i][1]);j++) {  /* and for each token in each match */
      response=get_objs_from_token(actor, obj, matches[i][1][j][0],matches[i][1][j][1], grammar[matches[i][0]]);
      if (sizeof(response)) {
        score+=response[0];
        if (matches[i][1][j][0]=="%file") {
          /* If it is '%file', fix things so the command gets a full path */
          matches[i][1][j][1]=this_object()->real_dir(matches[i][1][j][1]);
        } else if (matches[i][1][j][0]=="%multifile") {
          /* !!! */
        }
        this_result+= ({ ({ matches[i][1][j][0], response[1], matches[i][1][j][1] }) });
      } else {
        good=0;
      }
    }
    if (good) {
      /* Filter the result to remove potential matches with %inside and 
         %multiinside where the objects aren't in the container, and with
         %outside and %multioutside where they are. */
      for (j=0;j<sizeof(this_result);j++) {
        if (this_result[j][0]=="%inside" || this_result[j][0]=="%multiinside") {
          k=0;
          container=nil;
          while (k<sizeof(this_result)) {
            if (this_result[k][0]=="%container") {
              container=this_result[k][1];
              k=sizeof(this_result);
            }
            k++;
          }
          if (container) {
            cont_objs=({ });
            for (k=0; k<sizeof(this_result[j][1]); k++) {
              if (this_result[j][1][k]->is_inside(container[0])) {
                cont_objs+=({this_result[j][1][k]});
              }
            }
            this_result[j][1]=cont_objs;
            if (!sizeof(cont_objs)) {
              good=0;
            }
          } else {
            good=0;
          }
        } else if (this_result[j][0]=="%outside" || this_result[j][0]=="%multioutside") {
          k=0;
          container=nil;
          while (k<sizeof(this_result)) {
            if (this_result[k][0]=="%container") {
              container=this_result[k][1];
              k=sizeof(this_result);
            }
            k++;
          }
          if (container) {
            cont_objs=({ });
            for (k=0; k<sizeof(this_result[j][1]); k++) {
              if (!this_result[j][1][k]->is_inside(container[0])) {
                cont_objs+=({this_result[j][1][k]});
              }
            }
            this_result[j][1]=cont_objs;
            if (!sizeof(cont_objs)) {
              good=0;
            }
          } else {
            good=0;
          }
        }
      }
    }
    if (good) {
      result+=({ ({ i, score, this_result }) });
    }
  }
  return result;
}

int execute_one_command(string command) {
  /* Parse and execute a command.
     Split the command into words.  Search the grammar for lines that could
     match (ignoring the token types - just matching text for now).  Each line
     could match in several ways, so generate a mapping for each way (for
     example, "put the fish in the bowl in the box" can match "put %multiheld
     in %container" in two ways: %multiheld="the fish",%container="the bowl in
     the box", or %multiheld="the fish in the bowl",%container="the box".
     Each candidate is now reduced to a list of token:text pairs.  Using
     knowledge of the world, try to convert each piece of text into a list
     of objects that fits the token type.  Take the best match, assemble
     the parameters, and call the function.
     Returns 1 on success (decoded the command and called the function),
     0 on failure (couldn't understand the command, or couldn't find the
     function).
  */
  string *words;
  mixed *matches;
  mixed *obj_matches;
  int i, best, best_score;
  mixed *objects;
  mixed *text;
  int grammar_line;
  string fun;
  object obj;

  words=explode(command, " ")-({""});
  if (sizeof(words)) {
    matches=find_text_matches(words);
    /* We've now got the basic text matches.  Now to convert each piece of
       text into a list of objects that fits the token type, rejecting all
       matches that don't give any objects. */
    obj_matches=find_objects(this_object(), matches);
    /* Now take the one with the highest score and execute it */
    best=-1;
    best_score=-1;
    for (i=0;i<sizeof(obj_matches);i++) {
      if (obj_matches[i][1]>best_score) {
        best=i;
        best_score=obj_matches[i][1];
      }
    }
    if (best!=-1) {
      grammar_line=matches[obj_matches[best][0]][0];
      objects=({ });
      text=({ });
      for (i=0; i<sizeof(grammar[grammar_line][GRAMMAR_PARAMS]); i++) {
        if (grammar[grammar_line][GRAMMAR_PARAMS][i]>=0) {
          objects+=({obj_matches[best][2][grammar[grammar_line][GRAMMAR_PARAMS][i]][1]});
          text+=({obj_matches[best][2][grammar[grammar_line][GRAMMAR_PARAMS][i]][2]});
        } else {
          objects+=({ ({ }) });
          text+=({ "" });
        }
      }
      fun=grammar[grammar_line][GRAMMAR_FUNCTION];
      obj=grammar[grammar_line][GRAMMAR_OBJ];
      if (function_object(fun, obj)) {
        call_other(obj, fun, this_object(), words[0], objects+text...);
        return 1;
      } else { /* Couldn't find the function */
        return 0;
      }
    } else { /* Couldn't decode the command */
      return 0;
    }
  } else {
    return 0;
  }
}

int execute_command(string command) {
  string *commands;
  string *bits;
  string this_command;
  int i;
  int result;

  result=1;
  command=replace_string(command, "\\|", "\0");
  commands=explode(command, "|");
  for (i=0; i<sizeof(commands); i++) {
    commands[i]=replace_string(commands[i], "\0", "|");
    if (function_object("resolve_aliases", this_object())) {
      this_command=this_object()->resolve_aliases(commands[i]);
      if (this_command!=commands[i]) {
        result=result*execute_command(this_command);
      } else {
        result=result*execute_one_command(commands[i]);
      }
    } else {
      result=result*execute_one_command(commands[i]);
    }
  }
  return result;
}

string get_command_help(string subject, object actor) {
  /* "actor" is provided so the help_str functions can customize their
     results to the user asking for help. */
  object *objs;  /* all objects that define this verb */
  object *supplied_objs;  /* all objects that provide help for this verb */
  object this_obj;
  string *bits;
  string keyword;
  string verb;
  int i;
  string result;

  objs=({ });
  bits=explode(subject, "/");
  verb=bits[sizeof(bits)-1];
  if (sizeof(bits)>1) {
    keyword=bits[0];
  } else {
    keyword=nil;
  }
  for (i=0;i<sizeof(grammar);i++) {
    if (grammar[i][GRAMMAR_VERB]==verb) {
      objs=(objs-({grammar[i][GRAMMAR_OBJ]}))+({grammar[i][GRAMMAR_OBJ]});
    }
  }
  supplied_objs=({ });
  for (i=0;i<sizeof(objs);i++) {
    if ((!keyword || (objs[i]->help_keyword()==keyword)) && objs[i]->help_str(verb, actor)) {
      supplied_objs+=({objs[i]});
    }
  }
  if (sizeof(supplied_objs)==0) {
    return "No help available.";
  } else if (sizeof(supplied_objs)==1) {
    return supplied_objs[0]->help_str(verb, actor);
  } else {
    result="Choose one of:\n";
    for (i=0;i<sizeof(supplied_objs);i++) {
      result+="  "+supplied_objs[i]->help_keyword()+"/"+verb+"\n";
    }
    return result;
  }
}
