/* All physical objects should inherit this one.  It contains the interface
   to the universe object, which keeps track of what physical objects exist
   and where they are. */

#include <kernel/universe.h>
#include <kernel/physical.h>
#include <type.h>

inherit command "/kernel/lib/command";

mapping properties;

int is_physical() {
  return 1;
}

void set_property(string property, mixed value) {
  properties[property]=value;
}

mixed query_property(string property) {
  return properties[property];
}

mapping all_properties() {
  return properties;
}

void add_to_universe() {
  /* Add this_object() to the universe.  It will be floating in the void
     until it is put somewhere. */
  UNIV_OBJ->add_object(this_object());
}

void remove_from_universe() {
  UNIV_OBJ->remove_object(this_object());
}

void set_article(string article) {
  set_property(PROP_ARTICLE, article);
}

string query_article() {
  return query_property(PROP_ARTICLE);
}

void set_gender(int gender) {
  set_property(PROP_GENDER, gender);
}

int query_gender() {
  return query_property(PROP_GENDER);
}

string query_pronoun(int which) {
  string result;

  result=({ ({ "it", "it", "itself", "its" }),
         ({ "him", "he", "himself", "his" }),
         ({ "her", "she", "herself", "her" }) })[query_gender()][which];
  return result;
}

void set_name(string name) {
  if (!query_article()) {
    if (name[0]=='a' || name[0]=='e' || name[0]=='i' || name[0]=='o' ||
        name[0]=='u') {
      set_article("an");
    } else {
      set_article("a");
    }
  }
  set_property(PROP_NAME, name);
}

string query_name() {
  return query_property(PROP_NAME);
}

string query_pos_name() {
  if (query_property(PROP_POS_NAME)) {
    return query_property(PROP_POS_NAME);
  } else {
    return query_property(PROP_NAME)+"'s";
  }
}

string query_art_name() {
  if (query_article() && query_article()!="") {
    return query_article()+" "+query_name();
  } else {
    return query_name();
  }
}

string query_pos_art_name() {
  if (query_article() && query_article()!="") {
    return query_article()+" "+query_pos_name();
  } else {
    return query_pos_name();
  }
}

void set_description(string description) {
  set_property(PROP_DESC, description);
}

string query_description() {
  return query_property(PROP_DESC);
}

void set_ids(mixed ids) {
  if (typeof(ids)==T_ARRAY) {
    set_property(PROP_IDS, ids);
  } else {
    set_property(PROP_IDS, ({ ids }));
  }
}

string *query_ids() {
  return query_property(PROP_IDS);
}

int check_id(string id) {
  string *ids;
  int i;

  ids=query_ids();
  if (ids) {
    for (i=0;i<sizeof(ids);i++) {
      if (lower_case(ids[i])==lower_case(id)) {
        return 1;
      }
    }
  }
  return 0;
}

object *all_objects() {
  /* return an array containing every object in the universe */
  return UNIV_OBJ->query_all_objects();
}

object *near_objects() {
  /* return an array containing the objects close to this one */
  object *result;
  object *inv;

  result=UNIV_OBJ->shallow_inventory(this_object());
  if (!result) {
    result=({ });
  }
  inv=UNIV_OBJ->shallow_inventory(UNIV_OBJ->query_environment(this_object()));
  if (inv) {
    result+=inv;
  }
  return result;
}

void move(mixed destination, object actor) {
  UNIV_OBJ->move_object(this_object(), destination, actor);
}

object query_environment() {
  return UNIV_OBJ->query_environment(this_object());
}

object *query_shallow_inventory() {
  return UNIV_OBJ->shallow_inventory(this_object());
}

object *query_deep_inventory() {
  return UNIV_OBJ->deep_inventory(this_object());
}

void create() {
  command::create();
  properties=([ ]);
  add_to_universe();
  set_name("nothing");
  set_article(nil);
  set_description("It doesn't really exist at all...");
  set_gender(0);
}

int is_inside(object container) {
  return UNIV_OBJ->is_inside(this_object(), container);
}

void init(object actor) {
  string path;
  mixed *dir_info;

  ::init(actor);
  path=explode(object_name(this_object()), "#")[0];
  dir_info=get_dir(path);
  if (sizeof(dir_info[0]) && dir_info[1][0]==-2) {
    "/kernel/command/soul"->add_soul_dir(path, actor, 1);
  }
}

void uninit(object actor) {
  string path;
  mixed *dir_info;

  ::uninit(actor);
  path=explode(object_name(this_object()), "#")[0];
  dir_info=get_dir(path);
  if (sizeof(dir_info[0]) && dir_info[1][0]==-2) {
    "/kernel/command/soul"->remove_soul_dir(path, actor, 1);
  }
}
