#include <config.h>
#include <kernel/kernel.h>
#include <kernel/universe.h>
#include <kernel/physical.h>
#include <type.h>
#include <kernel/binary.h>

/* User object states for receive_message - what is the message? */
/* User is logged in.  Message is a command */
#define STATE_NORMAL 0
#define STATE_EDITING 1

inherit living "/kernel/lib/living";
inherit aliases "/kernel/lib/aliases";
inherit sprintf "/kernel/lib/sprintf";
inherit names "/kernel/lib/names";

static int state;
static int last_active;
static object *command_handlers;
static object editor;

string password;
static string new_password;
string current_directory;
int login_time, connected;

void create() {
  living::create();
  password="";
  set_name("Nobody");
  set_article("");
  set_ids("nobody");
  set_description("A shadow, not entirely of this world.");
  current_directory="/";
  last_active=time();
  editor=nil;
}

object connection_object() {
  return USER_MANAGER->query_connection(this_object());
}

int send_message(mixed arg) {
  if (connection_object()) {
    return connection_object()->send_message(arg);
  }
}

void prompt() {
  switch (state) {
    case STATE_NORMAL:
      send_message(1);
      send_message(">");
      break;
    case STATE_EDITING:
      send_message(1);
      send_message(":");
      break;
  }
}

void load_commands(string path) {
  mixed *files;
  int i;
  object obj;

  files=get_dir(path+"/*")[0];
  for (i=0;i<sizeof(files);i++) {
    if (lower_case(files[i][strlen(files[i])-2..strlen(files[i])-1])==".c") {
      obj=find_object(path+"/"+files[i]);
      if (!command_handlers) {
	command_handlers=({ });
      }
      command_handlers+=({ obj });
      call_other(obj, "init", this_object());
    }
  }
  depends_on(path);
}

void unload_commands() {
  /* This will unload commands defined by the command handlers that 
     load_commands loaded.  It won't touch commands defined by nearby
     objects. */
  int i;

  if (command_handlers) {
    for (i=0;i<sizeof(command_handlers);i++) {
      command_handlers[i]->uninit(this_object());
    }
  }
}

void prepare_to_die() {
  execute_command("logout");
  send_event(all_objects(), "logout", this_object());
  unload_commands();
  USER_MANAGER->remove_user(this_object());
}

void kill_me() {
  connection_object()->quitting();
  destruct_object(connection_object());
  destruct_object(this_object());
}

void save_user() {
  string filename;
  string name;

  name=query_name();
  filename=lower_case("/save/users/"+name[0..0]+"/"+name);
  save_object(filename);
}

void relogin() {
  /* They've just reconnected */
  connected=1;
  send_event(all_objects(), "relogin", this_user());
  state=STATE_NORMAL;
  prompt();
}

string query_home_dir() {
  return "/users/"+lower_case(query_name());
}

void load_command_dirs() {
  load_commands("/kernel/command");
  load_commands("/kernel/wiz_cmd");
  load_commands(query_home_dir()+"/command");
}

void login() {
  /* They've just succesfully logged in */
  connected=1;
  login_time=time();
  send_event(all_objects(), "login", this_user());
  load_command_dirs();
  move("/rooms/start_room", this_object());
  execute_command("login");
  state=STATE_NORMAL;
  prompt();
}

void setup_new_user(string name) {
  /* They're new! */
  connected=1;
  login_time=time();
  set_name(capitalize(name));
  set_article("");
  set_description(query_name()+" is just an ordinary person.");
  set_ids(lower_case(name));
  send_event(all_objects(), "login", this_user());
  load_command_dirs();
  set_alias("'_", "say");
  set_alias(":_", "emote");
  set_alias(";_", "emotes");
  set_alias("l", "look");
  set_alias("x", "examine");
  set_alias("i", "inv");
  set_alias("n", "north");
  set_alias("s", "south");
  set_alias("e", "east");
  set_alias("w", "west");
  set_alias("u", "up");
  set_alias("d", "down");
  send_message("\n");
  more_string(read_file("/data/newbie.txt"));
  send_message("\n");
  save_user();
  move("/rooms/start_room", this_object());
  execute_command("login");
  state=STATE_NORMAL;
  prompt();
}

string real_dir(string dir) {
  /* return the path to the directory.  Relative to current_directory if
     dir doesn't start with / 
     !!! going to have to fix this so it does .. */

  if (dir=="") {
    return current_directory;
  }
  if (dir[0..0]=="/") {
    return dir;
  } else {
    return current_directory+"/"+dir;
  }
}

void set_current_directory(string path) {
  current_directory=real_dir(path);
}

void set_password(string new_password) {
  if (this_user()==this_object() && explode(object_name(previous_object()), "/")[0]=="kernel") {
    password=new_password;
  }
}

string query_password() {
  return password;
}

int query_idle_time() {
  return time()-last_active;
}

void receive_message_normal(string message) {
  /* message is a command. */
  int result;

  if (message!="") {
    result=execute_command(message);
    if (!result) {
      send_message("What?\n");
    }
  }
}

void edit(string path) {
  if (!editor) {
    editor=find_object("/kernel/obj/editor");
    editor=clone_object(editor);
    editor->execute("e "+path);
  }
  state=STATE_EDITING;
}

void receive_message_editing(string message) {
  string response;

  if (editor) {
    response=editor->execute(message);
    if (response) {
    send_message(response);
    }
    if (!query_editor(editor)) {
      destruct_object(editor);
      state=STATE_NORMAL;
    }
  } else {
    state=STATE_NORMAL;
  }
}

void receive_message(string message) {
  if (connection_object()->query_echo()) {
    send_message(message+"\n");
  }
  last_active=time();
  switch(state) {
    case STATE_NORMAL:
      receive_message_normal(message);
      break;
    case STATE_EDITING:
      receive_message_editing(message);
      break;
  }
}

int query_cols() {
  return 79;
}

void event_say(object sender, string sender_name, string verb, string message) {
  if (message=="") {
    send_message(sprintf("%*i=s\n", query_cols(), -strlen(sender_name+" "+verb), sender_name+" "+verb+" nothing."));
  } else {
    send_message(sprintf("%*i*=s\n", query_cols(), -strlen(sender_name+" "+verb), sender_name+" "+verb+": "+message));
  }
}

void event_emote(object sender, string message) {
  send_message(message+"\n");
}

void event_login(object sender, object person) {
  if (person!=this_object()) {
    send_message(person->query_name()+" arrives.\n");
    connection_object()->send_binary_packet(BIN_USER_ON+person->query_name());
  }
}

void event_relogin(object sender, object person) {
  if (person!=this_object()) {
    send_message(person->query_name()+" returns.\n");
    connection_object()->send_binary_packet(BIN_USER_ON+person->query_name());
  }
}

void event_netdead(object sender, object person) {
  if (person!=this_object()) {
    send_message(person->query_name()+" goes net-dead.\n");
    connection_object()->send_binary_packet(BIN_USER_ON+person->query_name());
  }
}

void event_logout(object sender, object person) {
  if (person!=this_object()) {
    send_message(person->query_name()+" leaves.\n");
    connection_object()->send_binary_packet(BIN_USER_OFF+person->query_name());
  }
}

void event_action(object sender, string name, string action) {
  send_message(sprintf("%*=s\n", query_cols(), name+" "+action));
}

string query_description() {
  string result;
  object *inv, *carrying, *wearing;
  int i;

  result=::query_description();
  inv=query_shallow_inventory();
  carrying=({ });
  wearing=({ });
  for (i=0; i<sizeof(inv); i++) {
    switch (inv[i]->query_property(PROP_RELATION)) {
      case REL_HELD: carrying+=({ inv[i] }); break;
      case REL_WORN: wearing+=({ inv[i] }); break;
    }
  }
  if (sizeof(carrying)) {
    result+="\n"+query_name()+" is carrying "+name_list(carrying, nil, nil, 0, 0);
  }
  if (sizeof(wearing)) {
    result+="\n"+query_name()+" is wearing "+name_list(wearing, nil, nil, 0, 0);
  }
  return result;
}

mixed *query_user_info() {
  return ({ query_idle_time(), query_ip_number(this_object()), 
            login_time, connected });
}

void event_move(object x, object actor, object obj, object old_loc, object new_loc) {
  int new_loc_living;
  string message;

  if (actor && explode(object_name(actor), "#")[0]=="/kernel/obj/tunnel") {
    return;
  }
  if (new_loc) new_loc_living=new_loc->query_property(UNIV_PROP_PERSON);
  if (actor && new_loc==actor->query_environment() && old_loc==actor && obj) {
    /* actor has dropped obj */
    if (this_object()==actor) {
      message="You drop "+obj->query_art_name()+".\n";
    } else {
      message=capitalize(actor->query_art_name())+" drops "+obj->query_art_name()+".\n";
    }
  } else if (actor && new_loc==actor && old_loc!=actor && obj) {
    /* actor has taken obj */
    if (this_object()==actor) {
      message="You take "+obj->query_art_name()+".\n";
    } else {
      message=capitalize(actor->query_art_name())+" takes "+obj->query_art_name()+".\n";
    }
  } else if (new_loc_living && old_loc==actor && obj) {
    /* actor has given obj to new_loc */
    if (this_object()==actor) {
      message="You give "+obj->query_art_name()+" to "+new_loc->query_art_name()+".\n";
    } else if (this_object()==new_loc) {
      message=capitalize(actor->query_art_name())+" gives "+obj->query_art_name()+" to you.\n";
    } else {
      message=capitalize(actor->query_art_name())+" gives "+obj->query_art_name()+" to "+new_loc->query_art_name()+".\n";
    }
  } else if (actor && old_loc==actor && new_loc && obj) {
    /* actor has put obj in new_loc */
    if (this_object()==actor) {
      message="You put "+obj->query_art_name()+" into "+new_loc->query_art_name()+".\n";
    } else {
      message=capitalize(actor->query_art_name())+" puts "+obj->query_art_name()+" into "+new_loc->query_art_name()+".\n";
    }
  } else {
    if (actor && actor->query_art_name()) {
      message=actor->query_art_name();
    } else {
      message="Someone";
    }
    message+=" makes ";
    if (obj && obj->query_art_name()) {
      message+=obj->query_art_name();
    } else {
      message+="something";
    }
    message+=" move from ";
    if (old_loc && old_loc->query_art_name()) {
      message+=old_loc->query_art_name();
    } else {
      message+="somewhere";
    }
    message+=" to ";
    if (new_loc && new_loc->query_art_name()) {
      message+=new_loc->query_art_name();
    } else {
      message+="somewhere";
    }
    message+="\n";
    message="";
  }
  send_message(message);
  if (this_object()==obj && connection_object() && new_loc) {
    connection_object()->send_tunnel_info("room:"+new_loc->query_art_name()+"\n");
  }
}

void go_netdead() {
  connected=0;
  send_event(all_objects(), "netdead", this_user());
}

void restore_me(string path) {
  restore_object(path);
}

void call_input_to(object obj, string function, string message) {
  call_other(obj, function, message);
}

void upgraded() {
  /* We've been upgraded, which might mean that one of the command directories
     has changed.  So dump the commands loaded from them and re-load. */
  unload_commands();
  load_command_dirs();
}
