/* This is the universe handler object.  It keeps track of all physical
   objects and their spatial relationships. */

#include <kernel/universe.h>
#include <kernel/physical.h>
#include <type.h>

object *objects;

void create() {
  ::create();
  objects=({ });
}

void add_object(object obj) {
  objects+=({ obj });
  obj->set_property(UNIV_PROP_PARENT, nil);
  obj->set_property(UNIV_PROP_CHILDREN, ({ }));
  obj->set_property(UNIV_PROP_ROOM, 0);
  obj->set_property(UNIV_PROP_CONTAINER, 0);
  obj->set_property(UNIV_PROP_CLOSED, 0);
  obj->set_property(UNIV_PROP_PERSON, 0);
}

object *query_all_objects() {
  return objects;
}

object query_environment(object obj) {
  if (!obj) {
    /* If it isn't an object, it doesn't have an inventory. */
    return nil;
  }
  if (obj->query_property(UNIV_PROP_ROOM)) {
    /* The environment of a room is the room */
    return obj;
  }
  if (obj->query_property(UNIV_PROP_CONTAINER) && obj->query_property(UNIV_PROP_CLOSED)) {
    /* The environment of a closed container is the container */
    return obj;
  }
  /* An object that isn't in anything doesn't have an environment */
  if (!obj->query_property(UNIV_PROP_PARENT)) {
    return nil;
  }
  /* The environment of everything else is the environment of the object it
     is in */
  return query_environment(obj->query_property(UNIV_PROP_PARENT));
}

object *shallow_inventory(object obj) {
  /* Returns an array containing every object in obj, but not the objects
     in them. */
  if (!obj) {
    return nil;
  }
  return obj->query_property(UNIV_PROP_CHILDREN);
}

object *deep_inventory(object obj) {
  /* Returns an array containing every object in obj, and all the objects
     in them, and all the object in them, and... */
  object *result;
  object *inv;
  object *obj_inv;
  int i;

  if (!obj) {
    return nil;
  }
  inv=obj->query_property(UNIV_PROP_CHILDREN);
  result=inv;
  for (i=0;i<sizeof(inv);i++) {
    obj_inv=deep_inventory(inv[i]);
    if (obj_inv) {
      result+=obj_inv;
    }
  }
  return result;
}

void call_init(object obj) {
  object parent;
  object *scope;
  int i;

  parent=obj->query_property(UNIV_PROP_PARENT);
  if (parent) {
    scope=({ parent })+parent->query_property(UNIV_PROP_CHILDREN)-({ obj });
    for (i=0;i<sizeof(scope);i++) {
      scope[i]->init(obj);
      obj->init(scope[i]);
    }
  }
}

void call_uninit(object obj) {
  object parent;
  object *scope;
  int i;

  parent=obj->query_property(UNIV_PROP_PARENT);
  if (parent) {
    scope=({ parent })+parent->query_property(UNIV_PROP_CHILDREN)-({ obj });
    for (i=0;i<sizeof(scope);i++) {
      scope[i]->uninit(obj);
      obj->uninit(scope[i]);
    }
  }
}

void move_object(mixed obj, mixed destination, object actor) {
  object old_parent;
  object *siblings;
  object *informees;
  int i;

  if (typeof(destination)==T_STRING) {
    destination=find_object(destination);
  }
  old_parent=obj->query_property(UNIV_PROP_PARENT);
  informees=({ obj, old_parent });
  if (old_parent!=destination) { /* No point moving it to the same place... */
    call_uninit(obj);
    if (old_parent) {
      siblings=old_parent->query_property(UNIV_PROP_CHILDREN);
      informees+=siblings;
      siblings-=({ obj });
      old_parent->set_property(UNIV_PROP_CHILDREN, siblings);
    }
    obj->set_property(UNIV_PROP_PARENT, destination);
    if (destination) {
      siblings=destination->query_property(UNIV_PROP_CHILDREN);
      informees+=siblings;
      siblings+=({ obj });
      destination->set_property(UNIV_PROP_CHILDREN, siblings);
    }
    informees+=previous_object()->near_objects();
    informees+=obj->near_objects();
    if (old_parent) informees+=old_parent->near_objects();
    if (destination) informees+=destination->near_objects();
    informees+=({ destination });
    obj->set_property(PROP_RELATION, REL_IN);
    call_init(obj);
    informees-=({ nil });
    /* move event: actor, object, old location, new location */
    send_event(informees, "move", actor, obj, old_parent, destination);
  }
}

void remove_object(object obj) {
  object *inv;
  int i;

  /* Destroy everything in the object as well.  The other option is to move
     them all into the parent of this object, destroying them if this
     object doesn't have a parent. */
  inv=shallow_inventory(obj);
  for (i=0;i<sizeof(inv);i++) {
    destruct_object(inv[i]);
  }
  move_object(obj, nil, previous_object());  /* Remove it from its current parent */
  objects-=({ obj });   /* Remove it from the universe */
}

int is_inside(object obj, object container) {
  obj=obj->query_property(UNIV_PROP_PARENT);
  while (obj && obj!=container) {
    obj=obj->query_property(UNIV_PROP_PARENT);
  }
  if (obj) {
    return 1;
  } else {
    return 0;
  }
}
